#include "DivideSJIS.h"

namespace {
    const bool COMMAS_SHALL_DISAPPEAR_IF_PLACED_IN_QUOTES = true;
    const char CLEAR_BIT = 48; // 30h, "0" 
    const char COMMA_BYTE = 44; // 2Ch, "," 
    const char SET_BIT = 49; // 31h, "1"
    const char UNDERLINE_BYTE = 95; // 5Fh, "_" 
    const int ANSI_N_0TO9_MINUS_SIZE = 11;
    const int ANSI_N_0TO9_SIZE = 10;
    const int F_SIZE_1 = F_SIZE - 1;
    const int H_81 = 129;
    const int H_FC = 252;
    const int KANA_HANDAKUTEN = 223; // DFh 
    const int KANA_KUTEN = 161; // A1h 
    const int KANJI_SIZE = 2;
    const int NOT_RECOGNIZED = -1;
    const int UPPER_LIMIT_TO_FIELD_NUMBER = F_SIZE - 1;
    const std::string COMMA_S = ",";
    const std::string ZERO = "0";
    const unsigned char STANDARD_QUOTE_BYTE = 34; // 22h, "\"" 
}

char DOT_C_S = '.';
char MINUS_C_S = '-';
char ANSI_N_0TO9[] = "0123456789";
char ANSI_N_0TO9_MINUS[] = "0123456789-";

int leadingNumbers(std::string & f) {
    char full[F_SIZE];
    int count;

    full[F_SIZE_1] = '\0';
    strncpy(full, f.c_str(), F_SIZE_1);

    count = strspn(full, ANSI_N_0TO9);
    return count;
}

int positiveOrNegativeNumbers(std::string & f) {
    char first_character[2] = {0, 0};
    char full[F_SIZE];
    char part[F_SIZE];
    int count = 0;
    int m;

    full[F_SIZE_1] = '\0';
    strncpy(full, f.c_str(), F_SIZE_1);

// Checking if a number or minus appears. 
    first_character[0] = full[0];
    m = strspn(first_character, ANSI_N_0TO9_MINUS);
    if( m > 0) {
        count += 1;

        memcpy(part, &full[1], F_SIZE_1);

// Counting numbers 
        m = strspn(part, ANSI_N_0TO9);
        if( m > 0) {
            count += m;
        }

    } else {
// Counting numbers 
        m = strspn(full, ANSI_N_0TO9);
        if( m > 0) {
            count += m;
        }

    }

    return count;
}

int countingNonnumbers(std::string & f) {
    char full[F_SIZE];
    char tail;
    int count;

    full[F_SIZE_1] = '\0';
    strncpy(full, f.c_str(), F_SIZE_1);

    count = strcspn(full, ANSI_N_0TO9);

    tail = full[ count - 1 ];
    if( tail == MINUS_C_S ) {
        count -= 1;
    }

    return count;
}

int trailingNumbers(std::string & f) {
    char full[F_SIZE];
    char tail;
    char whole[F_SIZE];
    int count, part_length, whole_length;

    full[F_SIZE_1] = '\0';
    strncpy(full, f.c_str(), F_SIZE_1);

    strcpy(whole, full);
    whole_length = strlen(whole);

    count = strspn( strrev(full), ANSI_N_0TO9);

    part_length = whole_length - count;
    tail = whole[ part_length - 1 ];
    if( tail == MINUS_C_S ) {
        count += 1;
    }

    return count;
}

std::string extract_decimal(std::string & f) {
    char *p;
    char *trace;
    char c;
    char full[F_SIZE];
    char head[F_SIZE];
    char part[F_SIZE];
    char tail[F_SIZE];
    int count, k;
    std::string r;

    full[F_SIZE_1] = '\0';
    strncpy(full, f.c_str(), F_SIZE_1);

// Checking if a dot appears. 
    p = strchr(full, DOT_C_S);
    if( p == NULL ) {
        std::string q("0.0");
        r = q;
    } else {
        trace = p;

        k = 0;
        for(p = &full[0]; p != trace; p++) {
            part[k++] = *p;
        }

        part[k] = '\0';
        std::string q(part);
        count = trailingNumbers(q);
        if( count == 0 ) {
            head[0] = '0';
            head[1] = '\0';
        } else if( count == 1 ) {
            p = trace - count;
            c = *p;

            if( c == '-' ) {
                head[0] = '0';
                head[1] = '\0';
            } else {
                head[0] = c;
                head[1] = '\0';
            }

        } else {
            p = trace - count;
            memcpy(head, p, count);
            head[count] = '\0';
        }

        p = trace + 1;
        strcpy(tail, p);
        count = strspn(tail, ANSI_N_0TO9);
        if( count == 0 ) {
            tail[0] = '0';
            tail[1] = '\0';
        } else {
            tail[count] = '\0';
        }

        full[0] = '\0';
        strcat(full, head);
        strcat(full, ".");
        strcat(full, tail);

        std::string b(full);
        r = b;
    }

    return r;
}

std::string extract_integer(std::string & field) {
    using namespace std;

    char part[F_SIZE];
    int count, p, tailLength;
    int fieldLength = field.length();
    std::string r, tail;

    p = countingNonnumbers(field);
    if( p == fieldLength ) { // countingNonnumbers() cannot find any number. 
        r = ZERO;
    } else {
        tailLength = fieldLength - p;
        tail = field.substr(p, tailLength);
        count = positiveOrNegativeNumbers(tail);
        if( count == 1 ) {
            strcpy(part, tail.c_str() );
            if( part[0] == MINUS_C_S ) {
                r = ZERO;
            } else {
                r = tail.substr(0, count);
            }

        } else {
            r = tail.substr(0, count);
        }

    }

    return r;
}

std::string transparent(std::string & field) {
    return field;
}

DividingSJIS::~DividingSJIS() {
    delete a;
}

DividingSJIS::DividingSJIS() {
    using namespace std;

    string SEPARATOR_S(COMMA_S);

    separator_byte = (unsigned char)COMMA_BYTE;

    ASCII_C = UNDERLINE_BYTE;
    quote_byte = STANDARD_QUOTE_BYTE;
    remove_separators = COMMAS_SHALL_DISAPPEAR_IF_PLACED_IN_QUOTES;
    separator = SEPARATOR_S;

    CLEAR_BIT_C = CLEAR_BIT;
    SET_BIT_C = SET_BIT;
    a = new AsQueue();
}

DividingSJIS::DividingSJIS(unsigned char s) {
    using namespace std;

    stringstream os;
    os << s;

    string SEPARATOR_S( os.str() );

    separator_byte = s;

    ASCII_C = UNDERLINE_BYTE;
    quote_byte = STANDARD_QUOTE_BYTE;
    remove_separators = COMMAS_SHALL_DISAPPEAR_IF_PLACED_IN_QUOTES;
    separator = SEPARATOR_S;

    CLEAR_BIT_C = CLEAR_BIT;
    SET_BIT_C = SET_BIT;
    a = new AsQueue();
}

DividingSJIS::DividingSJIS(unsigned char q, unsigned char s, bool e) {
    using namespace std;

    stringstream os;
    os << s;

    string SEPARATOR_S( os.str() );

    quote_byte = q;
    remove_separators = e;
    separator_byte = s;

    ASCII_C = UNDERLINE_BYTE;
    separator = SEPARATOR_S;

    CLEAR_BIT_C = CLEAR_BIT;
    SET_BIT_C = SET_BIT;
    a = new AsQueue();
}

DividingSJIS::DividingSJIS(unsigned char q, unsigned char s, bool e, unsigned char ascii) {
    using namespace std;

    stringstream os;
    os << s;

    string SEPARATOR_S( os.str() );

    ASCII_C = (char)ascii;
    quote_byte = q;
    remove_separators = e;
    separator_byte = s;

    separator = SEPARATOR_S;

    CLEAR_BIT_C = CLEAR_BIT;
    SET_BIT_C = SET_BIT;
    a = new AsQueue();
}

int DividingSJIS::complementing_fields_list(char *buffer, int bufferLength, std::string & extra, unsigned char bond) {
    smashed_to_pieces(buffer, bufferLength, manyFields);

    vector_to_array_complementing(extra, bond);
    return field_count;
}

int DividingSJIS::editing_field(char *buffer, int bufferLength, int field_number) {
    smashed_to_pieces(buffer, bufferLength, manyFields);

    vector_to_array_editing(transparent, field_number);
    return field_count;
}

int DividingSJIS::extracting_decimal_from_field(char *buffer, int bufferLength, int field_number) {
    smashed_to_pieces(buffer, bufferLength, manyFields);

    vector_to_array_editing(extract_decimal, field_number);
    return field_count;
}

int DividingSJIS::extracting_integer_from_field(char *buffer, int bufferLength, int field_number) {
    smashed_to_pieces(buffer, bufferLength, manyFields);

    vector_to_array_editing(extract_integer, field_number);
    return field_count;
}

int DividingSJIS::fieldCount() const {
    return field_count;
}

int DividingSJIS::mapping_fields_list(char *buffer, int bufferLength) {
    smashed_to_pieces(buffer, bufferLength, manyFields);

    vector_to_array_mapping();
    return field_count;
}

int DividingSJIS::reversing_fields_list(char *buffer, int bufferLength) {
    smashed_to_pieces(buffer, bufferLength, manyFields);

    vector_to_array_arranging_in_reverse_order();
    return field_count;
}

int DividingSJIS::search_separator(char *buffer, int bufferLength, int beginning) {
    bool no_splitting = false;
    int r = NOT_RECOGNIZED;
    unsigned char n;

    for(int k = beginning; k < bufferLength; ++k) {
        n = (unsigned char)buffer[k];

        if( no_splitting ) {
            if( n == quote_byte) {
                no_splitting = false;
            }

            continue; // Commas disappear if placed in quotes. 
        }

// splitting 
        if( n == quote_byte) {
            no_splitting = true;
        } else if( n == separator_byte ) {
            r = k;
            break;
        }

    }

    return r;
}

int DividingSJIS::smashed_to_pieces(char *buffer, int bufferLength, bool *selected) {
    using namespace std;

    bool kanji = false;
    bool no_splitting = false;
    char c;
    char first_byte = COMMA_BYTE;
    int b = 0;
    int field_number = 1;
    string *p = new string("");
    unsigned char n;

    for(int k = 0; k < bufferLength; ++k) {
        c = buffer[k];
        n = (unsigned char)c;

        if( kanji ) {
            if( selected[field_number] ) {
                dough[b] = first_byte;
                dough[ b + 1 ] = c;
                b += KANJI_SIZE;
            }

            kanji = false;
            continue;
        }

        if( H_81 <= n && n <= H_FC ) {
            if( KANA_KUTEN <= n && n <= KANA_HANDAKUTEN ) {
                if( selected[field_number] ) {
                    dough[b++] = c;
                }

                kanji = false;
            } else {
                first_byte = c;
                kanji = true;
            }

            continue;
        }

// ASCII 
        if( no_splitting ) {
            if( n == quote_byte ) {
                no_splitting = false;
                if( selected[field_number] ) {
                    dough[b++] = c;
                }

            } else if( n == separator_byte ) {
                if( remove_separators ) {
                    continue; // Commas disappear if placed in quotes. 
                } else {
                    if( selected[field_number] ) {
                        dough[b++] = c;
                    }

                }

            } else {
                if( selected[field_number] ) {
                    dough[b++] = c;
                }

            }

            continue;
        }

// splitting 
        if( n == quote_byte ) {
            no_splitting = true;
            if( selected[field_number] ) {
                dough[b++] = c;
            }

        } else if( n == separator_byte ) {
            if( selected[field_number] ) {
                dough[b] = '\0';
                doughLength = b;
                delete p;
                p = new string(dough);
                a->push_back(*p); // AsQueue
            }

            b = 0;
            field_number += 1;
            if( field_number > field_count ) {
                break;
            }

        } else {
            if( selected[field_number] ) {
                dough[b++] = c;
            }

        }

    }

    if( b > 0 ) {
        dough[b] = '\0';
        doughLength = b;
        delete p;
        p = new string(dough);
        a->push_back(*p); // AsQueue
    } else {
        field_number -= 1;
    }

    field_count = field_number;
    delete p;
    return b;
}

int DividingSJIS::unitedSize() const {
    return size_of_united_items;
}

std::string DividingSJIS::reverse_unite(char *buffer, int bufferLength) {
    using namespace std;

    int item_count;
    stringstream os;

    reversing_fields_list(buffer, bufferLength);

    item_count = f->itemCount();
    for(int n = 0; n < item_count; ++n) {
        os << f->item(n) << separator;
    }

    combo = os.str();
    return combo;
}

std::string DividingSJIS::supplying_bits(int c) {
    using namespace std;

    stringstream os;

    field_count = c;
    if( field_count > UPPER_LIMIT_TO_FIELD_NUMBER ) {
        field_count = UPPER_LIMIT_TO_FIELD_NUMBER;
    }

    for(int i = 0; i < field_count; ++i) {
        os << SET_BIT_C;
    }

    return os.str();
}

std::string DividingSJIS::surge_of_set_bits(int c, int interval_0, int interval_1) {
    using namespace std;

    int bodyLength, headLength, tailLength;
    stringstream os;

/*
  surge_of_set_bits
0011111000000000000, for example
Setting bits from interval_0 to interval_1
*/

    field_count = c;
    if( interval_1 > field_count ) {
        interval_1 = field_count;
    }

    headLength = interval_0 - 1;
    for(int i = 0; i < headLength; ++i) {
        os << CLEAR_BIT_C;
    }

    bodyLength = interval_1 - interval_0 + 1;
    for(int i = 0; i < bodyLength; ++i) {
        os << SET_BIT_C;
    }

    tailLength = field_count - interval_1;
    for(int i = 0; i < tailLength; ++i) {
        os << CLEAR_BIT_C;
    }

    return os.str();
}

std::string DividingSJIS::unite(AsQueue *a, UniversalComboAppend *c, char *buffer, int bufferLength) {
    using namespace std;

    int item_count;
    string g;

    if( a->size() == 0 ) {
        mapping_fields_list(buffer, bufferLength);
    }

    item_count = f->itemCount();
    for(int n = 0; n < item_count; ++n) {
        g = f->item(n);
        c->append(g);
        c->append(separator);
    }

    size_of_united_items = c->comboSize();
    return c->built();
}

std::string DividingSJIS::unite(char *buffer, int bufferLength) {
    using namespace std;

    stringstream os;

    smashed_to_pieces(buffer, bufferLength, manyFields);

    for(AsQueue::iterator i = a->begin(); i != a->end(); ++i) {
        os << *i << separator;
    }

    combo = os.str();
    size_of_united_items = combo.length();
    return combo;
}

std::string DividingSJIS::unite(char *buffer, int bufferLength, bool listed) {
    using namespace std;

    int item_count;
    stringstream os;

    if(! listed ) {
        mapping_fields_list(buffer, bufferLength);
    }

	item_count = f->itemCount();
    for(int n = 0; n < item_count; ++n) {
        os << f->item(n) << separator;
    }

    combo = os.str();
    size_of_united_items = combo.length();
    return combo;
}

void DividingSJIS::vector_to_array_arranging_in_reverse_order() {
    using namespace std;

    int k = a->size() - 1;

    for(AsQueue::iterator i = a->begin(); i != a->end(); ++i) {
        f->setting(*i, k);
        k -= 1;
    }

}

void DividingSJIS::vector_to_array_complementing(std::string & extra, unsigned char bond) {
    using namespace std;

    int a_count = 0;
    string s;
    stringstream *os = new stringstream();

    for(AsQueue::iterator i = a->begin(); i != a->end(); ++i) {
        delete os;
        os = new stringstream();
        *os << *i << bond << extra;
        s = os->str();
        f->setting(s, a_count);
        a_count += 1;
    }

    delete os;
}

void DividingSJIS::vector_to_array_editing(EditFunction editing, int field_number) {
    using namespace std;

    int a_count = 0;
    int field_number_1 = field_number - 1;
    string s;

    for(AsQueue::iterator i = a->begin(); i != a->end(); ++i) {
        if( a_count == field_number_1 ) {
            s = editing(*i);
            f->setting(s, a_count);
        } else {
            f->setting(*i, a_count);
        }

        a_count += 1;
    }

}

void DividingSJIS::vector_to_array_mapping() {
    using namespace std;

    int a_count = 0;

    for(AsQueue::iterator i = a->begin(); i != a->end(); ++i) {
        f->setting(*i, a_count);
        a_count += 1;
    }

}

unsigned char DividingSJIS::quoteCode() const {
    return quote_byte;
}

unsigned char DividingSJIS::separatorCode() const {
    return separator_byte;
}

void DividingSJIS::setting(FieldsList *fields_list, int c) {
    f = fields_list;
    field_count = c;

    manyFields[0] = true;
    for(int m = 1; m <= field_count; ++m) {
        manyFields[m] = true;
    }

}

void DividingSJIS::setting(FieldsList *fields_list, int c, char *pn) {
    int k;

    f = fields_list;
    field_count = c;

    manyFields[0] = true;
    for(int m = 1; m <= field_count; ++m) {
        k = m - 1;
        pattern[k] = pn[k];

        if( pattern[k] == SET_BIT ) {
            manyFields[m] = true;
        } else {
            manyFields[m] = false;
        }

    }

}

