#ifndef HYPHEN_TERM_H
#define HYPHEN_TERM_H

#define F_SIZE 3072
#define P_SIZE 128 // P_SIZE > Q_SIZE
#include <cmath>
#include <fstream>
#include <mem.h>
#include "Numbered.h"
#include "parameters.h"

typedef struct Triplet Triplet;
struct Triplet {
    int t[3];
};

class HyphenTerms {
    bool binary_represented;
    bool deniable[9];
    char CLEAR_BIT_C, GAP_C, SET_BIT_C, TAB_C;
    char buffer[F_SIZE];
    char c[C_SIZE];
    int bar_estimate, buffer_size, console_height, console_width, edge_estimate, error_level, field_count, focal_point, hidden_pane, interval_from, interval_to, which_corner, waiting_time;
    int byteCode[5];
    int figure[13];
    int maximum[3];
    int minimum[3];
    ParameterList *listing;
    std::string ASCII, BYTES, EBB, FLOOD, exe_name, joint_piece, label_piece, NEW_LINE, path_name, pattern, quote_piece, quoted_operand, separator_piece;
// ASCII := ASCII_1 + '"';
    std::string acquisition[Q_SIZE];
    std::string bit_pattern_array[3];
    std::string commandParameter[P_SIZE];
    std::string list_name[3];
    std::string piece[5];
    Triplet triplet_in_context;
    unsigned char joint_byte, label_byte, quote_byte, separator_byte;
public:
    ~HyphenTerms();
    HyphenTerms(ParameterList *L); // presetting

    bool extensionEqualsCSV(std::string & file_name);
    bool guideFile();
    bool jointPieces(std::string & s);
    bool killCMD();
    bool labelPieces(std::string & s);
    bool notAutomatically();
    bool quotePieces(std::string & s);
    bool separatorPieces(std::string & s);
    bool showWhere();
    bool startHelp();
    bool titleVersion();
    bool versionNumber();
    int acquireFromText();
    int atPoint();
    int barEstimate();
    int bufferSize();
    int consoleHeight();
    int consoleWidth();
    int copyingStringToArray(char *s, int s_length);
    int edgeEstimate();
    int errorLevel();
    int hexadecimal_to_integer(std::string & s);
    int hiddenPane();
    int tripletAfter();
    int tripletBefore();
    int tripletPoint();
    int waitingTime();
    int whichCorner();
    std::string exeName();
    std::string jointPiece();
    std::string joints();
    std::string labelPiece();
    std::string labels();
    std::string listName(int index);
    std::string multipleBitPattern();
    std::string negativePattern();
    std::string pathName();
    std::string quotedOperand();
    std::string quotePiece();
    std::string quotes();
    std::string separatorPiece();
    std::string separators();
    std::string singleSurge(); // not Triplet 
    std::string surge_of_set_bits(int interval_0, int interval_1);
    Triplet headEnd();
    Triplet inContext();
    Triplet nearPoint();
    Triplet tailEnd();
    Triplet tailEnd(int line_count);
    unsigned char jointByte();
    unsigned char jointCode();
    unsigned char labelByte();
    unsigned char labelCode();
    unsigned char punctuation(std::string & s, int q);
    unsigned char quoteByte();
    unsigned char quoteCode();
    unsigned char separatorByte();
    unsigned char separatorCode();
    void setting();
};

#endif // HYPHEN_TERM_H
